/* Copyright (C) 2005 by George Williams */
/*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.

 * The name of the author may not be used to endorse or promote products
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "pfaedit.h"
#include <stdio.h>
#include <math.h>
#include "splinefont.h"
#include "views.h"
#include "stemdb.h"

void DoMF(SplineChar *sc,struct mfdata *mfd) {
    struct glyphdata *gd;

    gd = GlyphDataBuild(sc,false);
    if ( gd==NULL )
return;
    if ( !mfd->centerout ) {
	_SplineCharAutoHint(sc,&mfd->bd,gd);
	    /* We want all the vertical stems */
	AddDiagonalStemsAsVertical(sc,gd);
	    /* Pretend diagonal stems are vertical to get counters for "K", etc */
	SCFigureVerticalCounterMasks(sc);
	    /* Find all sets of counters & stems */
	BuildHorizontalMappingCountersStems(sc,mfd);
	    /* Pick one counter/stem hint, and use it to establish a horizontal*/
	    /* map. Expand the counters and the stems and see where they should*/
	    /* go */
	BuildVerticalMappingZones(sc,mfd);
	    /* Vertical zones have different constraints. The top of "O" should*/
	    /* only expand down, and the bottom should only expand up. */
    } else {
	BuildSimpleHVZones(sc,mfd);
	    /* Here the center lines of stems remain at the same place... */
	    /* well we will allow changes of left/right side barrings, but */
	    /* no internal expansion/contraction */
    }
    PositionStemCenterLineEndPoints(mfd,gd);
	/* Figure out where the centerline of each stem should go */
	/*  (apply the above transforms) */
    PositionStemPoints(mfd,gd);
	/* For each point on each real stem figure where it goes */
	/* Simplest case is just expanding out from the center */
	/* For points on two stems, we must intersect the stem edges */
    InterpolateInternalPoints(mfd,gd);
	/* Points which aren't on real stems get interpolated between */
	/* any positioned points around them. If no positioned points */
	/* then apply the bare zone mappings */
    InterpolateControlPoints(mfd,gd);
    SaveStateAndBackground(sc);
    RegenerateSplines(sc,mfd,gd);
}
