#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define true 1
#define false 0

#define copy	strdup
#define copyn	strndup

typedef struct growbuf {
    char *pt;
    char *base;
    char *end;
} GrowBuf;

static void GrowBuffer(GrowBuf *gb,int len) {
    if ( len<400 ) len = 400;
    if ( gb->base==NULL ) {
	gb->base = gb->pt = malloc(len);
	gb->end = gb->base + len;
    } else {
	int off = gb->pt-gb->base;
	len += (gb->end-gb->base);
	gb->base = realloc(gb->base,len);
	gb->end = gb->base + len;
	gb->pt = gb->base+off;
    }
}

static void GBAddChar(GrowBuf *gb,int ch) {
    if ( gb->pt+1 >= gb->end )
	GrowBuffer(gb,1);
    *(gb->pt++) = ch;
}

static void GBSlurpString(FILE *file,GrowBuf *gb) {
    /* We've read the left paren. Search for a matching right paren */
    int paren_cnt = 1;
    int ch, saw_back=0;

    while ((ch=getc(file))!=EOF ) {
	GBAddChar(gb,ch);
	if ( saw_back )
	    saw_back = false;
	else if ( ch=='\\' )
	    saw_back = true;
	else if ( ch=='(' )
	    ++paren_cnt;
	else if ( ch==')' ) {
	    if ( --paren_cnt == 0 )
    break;
	}
    }
    GBAddChar(gb,'\0');
}

static void GBSlurpProc(FILE *file,GrowBuf *gb) {
    int curly_cnt = 1;
    int ch;

    while ((ch=getc(file))!=EOF ) {
	GBAddChar(gb,ch);
	if ( ch=='%' ) {
	    while ( (ch=getc(file))!=EOF ) {
		GBAddChar(gb,ch);
		if ( ch=='\n' || ch=='\r' )
	    break;
	    }
	} else if ( ch=='(' )
	    GBSlurpString(file,gb);
	else if ( ch=='{' )
	    ++curly_cnt;
	else if ( ch=='}' ) {
	    if ( --curly_cnt == 0 )
    break;
	}
    }

    while ( isspace(ch=getc(file)) )
	GBAddChar(gb,ch);
    GBAddChar(gb,'\0');
    if ( ch!=EOF )
	ungetc(ch,file);
}

static void GBSlurpArray(FILE *file,GrowBuf *gb) {
    int bracket_cnt = 1;
    int ch;

    while ((ch=getc(file))!=EOF ) {
	GBAddChar(gb,ch);
	if ( ch=='%' ) {
	    while ( (ch=getc(file))!=EOF ) {
		GBAddChar(gb,ch);
		if ( ch=='\n' || ch=='\r' )
	    break;
	    }
	} else if ( ch=='(' )
	    GBSlurpString(file,gb);
	else if ( ch=='{' ) {
	    GBSlurpProc(file,gb);
	    --gb->pt;
	} else if ( ch=='[' )
	    ++bracket_cnt;
	else if ( ch==']' ) {
	    if ( --bracket_cnt == 0 )
    break;
	}
    }

    while ( isspace(ch=getc(file)) )
	GBAddChar(gb,ch);
    GBAddChar(gb,'\0');
    if ( ch!=EOF )
	ungetc(ch,file);
}

static int GBSlurpToArray(FILE *file,GrowBuf *gb) {
    int ch;

    while ((ch=getc(file))!=EOF ) {
	if ( ch=='[' ) {
	    GBAddChar(gb,'\0');
return( true );
	}
	GBAddChar(gb,ch);
	if ( ch=='%' ) {
	    while ( (ch=getc(file))!=EOF ) {
		GBAddChar(gb,ch);
		if ( ch=='\n' || ch=='\r' )
	    break;
	    }
	} else if ( ch=='{' ) {
	    GBAddChar(gb,'\0');
return( false );
	}
    }
    GBAddChar(gb,'\0');
return( -1 );
}

static void GBSlurpToProc(FILE *file,GrowBuf *gb) {
    int ch;

    while ((ch=getc(file))!=EOF ) {
	GBAddChar(gb,ch);
	if ( ch=='%' ) {
	    while ( (ch=getc(file))!=EOF ) {
		GBAddChar(gb,ch);
		if ( ch=='\n' || ch=='\r' )
	    break;
	    }
	} else if ( ch=='{' || ch=='[' )
    break;
    }
    GBAddChar(gb,'\0');
}

static char *ConcatStrings(char *str1, char *str2, char *str3) {
    int len=0;
    char *ret;

    if ( str2==NULL && str3==NULL )
return( str1 );
    if ( str1==NULL && str3==NULL )
return( str2 );
    if ( str1==NULL && str2==NULL )
return( str3 );

    if ( str1!=NULL )
	len = strlen(str1);
    if ( str2!=NULL )
	len += strlen(str2);
    if ( str3!=NULL )
	len += strlen(str3);
    ret = malloc(len+1);
    *ret = '\0';
    if ( str1!=NULL )
	strcpy(ret,str1);
    if ( str2!=NULL )
	strcat(ret,str2);
    if ( str3!=NULL )
	strcat(ret,str3);
    free(str1); free(str2); free(str3);
return( ret );
}

static char **ParseOtherSubrs(FILE *file) {
    GrowBuf gb;
    char *copyright=NULL;
    char *osubrs[20];
    char *before=NULL;
    int i, ret;
    char **result = NULL;
    char *pt;

    memset(osubrs,'\0',sizeof(osubrs));
    memset(&gb,'\0',sizeof(GrowBuf));
    GrowBuffer(&gb,1);			/* Force allocation so I don't need to check for an empty buffer later */

    ret = GBSlurpToArray(file,&gb);
    if ( ret==-1 ) {
	free(gb.base);
return( NULL );
    }

    copyright = copy(gb.base);
    for ( i=0; i<sizeof(osubrs)/sizeof(osubrs[0]); ++i ) {
	/* we only support reading in 14 other subrs. That's as many as we */
	/*  understand. There are some more but they are for mm fonts and */
	/*  we must generate them depending on the number of axes, etc. */
	before = NULL;
	if ( ret ) {		/* Only could be false first time round */
	    gb.pt = gb.base;
	    GBSlurpToProc(file,&gb);
	    for ( pt = gb.base; isspace(*pt); ++pt);
	    if ( strncmp(pt,"ifelse",6)==0 && i>=2 ) {
		osubrs[i-2] = ConcatStrings(osubrs[i-2],osubrs[i-1],copy("ifelse"));
		osubrs[--i] = NULL;
		pt += 6;
	    } else if ( pt[0]=='i' && pt[1]=='f' && pt[2]!='e' && i>=1 ) {
		osubrs[i-1] = ConcatStrings(osubrs[i-1],NULL,copy("if"));
		pt += 2;
	    }
	    gb.pt = pt;
	    while ( isspace( *pt ))
		++pt;
	    if ( strncmp(pt,"executeonly",11)==0 ) {
		pt += 11;
		while ( isspace( *pt )) ++pt;
	    }
	    if ( pt!=gb.pt && i>0 )
		osubrs[i-1] = ConcatStrings(osubrs[i-1],NULL,copyn(gb.pt,pt-gb.pt));
	    else
		pt = gb.pt;
	    if ( *pt=='[' ) {
		gb.pt = gb.base;
		GBSlurpArray(file,&gb);
		osubrs[i-1] = ConcatStrings(osubrs[i-1],copy("["),copy(gb.base));
		--i;
    continue;
	    } else if ( *pt!='\0' )
		before = copy(pt);
	} else {
	    before = copy("{");
	    ret = true;
	}
	gb.pt = gb.base;
	GBSlurpProc(file,&gb);
	if ( feof(file)) {
	    free(before);
    break;
	}
	osubrs[i] = ConcatStrings(before,copy(gb.base),NULL);
    }

    while ( i<13 ) {
	osubrs[i] = copy(" {}\n");
	++i;
    }
    if ( i==13 ) {
	osubrs[i] = copy(" {2 {cvi {{pop 0 lt {exit} if} loop} repeat} repeat}\n");
	++i;
    }

    result = malloc(16*sizeof(char *));
    result[0] = copyright;
    for ( i=0; i<14; ++i )
	result[i+1] = osubrs[i];
    result[15] = NULL;

    while ( i<sizeof(osubrs)/sizeof(osubrs[0]) ) {
	free(osubrs[i]);
	++i;
    }
    free(gb.base);
return( result );
}

int main(int argc, char **argv) {
    char **result;
    int i;

    result = ParseOtherSubrs(stdin);
    if ( result==NULL ) {
	fprintf( stderr,"Failure\n" );
	exit(1);
    } else {
	for ( i=0; result[i]!=NULL; ++i ) {
	    fputs(result[i],stdout);
	    printf("\n\n\n");
	}
    }
return( 0 );
}
